from os import getenv as os_getenv
import sys
import textwrap
from typing import Dict, Optional

import colorama

# Colorama is needed for colored logs on Windows because we're using logger.info
# intead of print(). If the Windows env doesn't have a TERM var set or it is set to None
# (i.e. in the case of Git Bash on Windows- this emulates Unix), then it's safe to initialize
# Colorama with wrapping turned on which allows us to strip ANSI sequences from stdout.
# You can safely initialize Colorama for any OS and the coloring stays the same except
# when piped to another process for Linux and MacOS, then it loses the coloring. To combat
# that, we will just initialize Colorama when needed on Windows using a non-Unix terminal.

if sys.platform == "win32" and (not os_getenv("TERM") or os_getenv("TERM") == "None"):
    colorama.init(wrap=True)

COLORS: Dict[str, str] = {
    "red": colorama.Fore.RED,
    "green": colorama.Fore.GREEN,
    "yellow": colorama.Fore.YELLOW,
    "reset_all": colorama.Style.RESET_ALL,
}


COLOR_FG_RED = COLORS["red"]
COLOR_FG_GREEN = COLORS["green"]
COLOR_FG_YELLOW = COLORS["yellow"]
COLOR_RESET_ALL = COLORS["reset_all"]


USE_COLOR = True
PRINTER_WIDTH = 80


def color(text: str, color_code: str) -> str:
    if USE_COLOR:
        return "{}{}{}".format(color_code, text, COLOR_RESET_ALL)
    else:
        return text


def printer_width() -> int:
    return PRINTER_WIDTH


def green(text: str) -> str:
    return color(text, COLOR_FG_GREEN)


def yellow(text: str) -> str:
    return color(text, COLOR_FG_YELLOW)


def red(text: str) -> str:
    return color(text, COLOR_FG_RED)


def line_wrap_message(msg: str, subtract: int = 0, dedent: bool = True, prefix: str = "") -> str:
    """Line wrap a message to a given printer width.

    Line wrap the given message to PRINTER_WIDTH - {subtract}. Convert double
    newlines to newlines and avoid calling textwrap.fill() on them (like
    markdown)
    """
    width = printer_width() - subtract
    if dedent:
        msg = textwrap.dedent(msg)

    if prefix:
        msg = f"{prefix}{msg}"

    # If the input had an explicit double newline, we want to preserve that
    # (we'll turn it into a single line soon). Support windows, too.
    splitter = "\r\n\r\n" if "\r\n\r\n" in msg else "\n\n"
    chunks = msg.split(splitter)
    return "\n".join(textwrap.fill(chunk, width=width, break_on_hyphens=False) for chunk in chunks)


def warning_tag(msg: str, event_name: Optional[str] = None) -> str:
    tag = f'[{yellow("WARNING")}]'
    if event_name:
        tag += f"[{event_name}]"
    return f"{tag}: {msg}"


def deprecation_tag(msg: str, event_name: Optional[str] = None) -> str:
    return warning_tag(f"Deprecated functionality\n\n{msg}", event_name)


def error_tag(msg: str, event_name: Optional[str] = None) -> str:
    tag = f'[{red("ERROR")}]'
    if event_name:
        tag += f"[{event_name}]"
    return f"{tag}: {msg}"



import dataclasses
from typing import Any, TypeVar

_R = TypeVar("_R", bound="Replaceable")


# TODO: remove from dbt_common.contracts.util:: Replaceable + references
class Replaceable:
    def replace(self: _R, **kwargs: Any) -> _R:
        return dataclasses.replace(self, **kwargs)  # type: ignore


_M = TypeVar("_M", bound="Mergeable")


class Mergeable(Replaceable):
    def merged(self: _M, *args: Any) -> _M:
        """Perform a shallow merge, where the last non-None write wins. This is
        intended to merge dataclasses that are a collection of optional values.
        """
        replacements = {}
        cls = type(self)
        for arg in args:
            for field in dataclasses.fields(cls):  # type: ignore
                value = getattr(arg, field.name)
                if value is not None:
                    replacements[field.name] = value

        return self.replace(**replacements)



from _typeshed import Incomplete
from enum import Enum
from mashumaro.config import BaseConfig as BaseConfig
from mashumaro.helper import pass_through as pass_through
from mashumaro.jsonschema.dialects import (
    DRAFT_2020_12 as DRAFT_2020_12,
    JSONSchemaDialect as JSONSchemaDialect,
)
from mashumaro.mixins.json import DataClassJSONMixin as DataClassJSONMixin
from typing import Any, Dict, List, Optional, Set, Union
from typing_extensions import TypeAlias

Number: TypeAlias = Union[int, float]
Null = object()

class JSONSchemaInstanceType(Enum):
    NULL: str
    BOOLEAN: str
    OBJECT: str
    ARRAY: str
    NUMBER: str
    STRING: str
    INTEGER: str

class JSONSchemaInstanceFormat(Enum): ...

class JSONSchemaStringFormat(JSONSchemaInstanceFormat):
    DATETIME: str
    DATE: str
    TIME: str
    DURATION: str
    EMAIL: str
    IDN_EMAIL: str
    HOSTNAME: str
    IDN_HOSTNAME: str
    IPV4ADDRESS: str
    IPV6ADDRESS: str
    URI: str
    URI_REFERENCE: str
    IRI: str
    IRI_REFERENCE: str
    UUID: str
    URI_TEMPLATE: str
    JSON_POINTER: str
    RELATIVE_JSON_POINTER: str
    REGEX: str

class JSONSchemaInstanceFormatExtension(JSONSchemaInstanceFormat):
    TIMEDELTA: str
    TIME_ZONE: str
    IPV4NETWORK: str
    IPV6NETWORK: str
    IPV4INTERFACE: str
    IPV6INTERFACE: str
    DECIMAL: str
    FRACTION: str
    BASE64: str
    PATH: str

DATETIME_FORMATS: Incomplete
IPADDRESS_FORMATS: Incomplete

class JSONSchema(DataClassJSONMixin):
    schema: Optional[str]
    type: Optional[JSONSchemaInstanceType]
    enum: Optional[List[Any]]
    const: Optional[Any]
    format: Optional[
        Union[JSONSchemaInstanceFormat, JSONSchemaStringFormat, JSONSchemaInstanceFormatExtension]
    ]
    title: Optional[str]
    description: Optional[str]
    anyOf: Optional[List["JSONSchema"]]
    reference: Optional[str]
    definitions: Optional[Dict[str, "JSONSchema"]]
    default: Optional[Any]
    deprecated: Optional[bool]
    examples: Optional[List[Any]]
    properties: Optional[Dict[str, "JSONSchema"]]
    patternProperties: Optional[Dict[str, "JSONSchema"]]
    additionalProperties: Union["JSONSchema", bool, None]
    propertyNames: Optional["JSONSchema"]
    prefixItems: Optional[List["JSONSchema"]]
    items: Optional["JSONSchema"]
    contains: Optional["JSONSchema"]
    multipleOf: Optional[Number]
    maximum: Optional[Number]
    exclusiveMaximum: Optional[Number]
    minimum: Optional[Number]
    exclusiveMinimum: Optional[Number]
    maxLength: Optional[int]
    minLength: Optional[int]
    pattern: Optional[str]
    maxItems: Optional[int]
    minItems: Optional[int]
    uniqueItems: Optional[bool]
    maxContains: Optional[int]
    minContains: Optional[int]
    maxProperties: Optional[int]
    minProperties: Optional[int]
    required: Optional[List[str]]
    dependentRequired: Optional[Dict[str, Set[str]]]

    class Config(BaseConfig):
        omit_none: bool
        serialize_by_alias: bool
        aliases: Incomplete
        serialization_strategy: Incomplete
    def __pre_serialize__(self, context: Optional[Dict]) -> JSONSchema: ...
    def __post_serialize__(self, d: Dict[Any, Any], context: Optional[Dict]) -> Dict[Any, Any]: ...
    def __init__(
        self,
        schema,
        type,
        enum,
        const,
        format,
        title,
        description,
        anyOf,
        reference,
        definitions,
        default,
        deprecated,
        examples,
        properties,
        patternProperties,
        additionalProperties,
        propertyNames,
        prefixItems,
        items,
        contains,
        multipleOf,
        maximum,
        exclusiveMaximum,
        minimum,
        exclusiveMinimum,
        maxLength,
        minLength,
        pattern,
        maxItems,
        minItems,
        uniqueItems,
        maxContains,
        minContains,
        maxProperties,
        minProperties,
        required,
        dependentRequired,
    ) -> None: ...

class JSONObjectSchema(JSONSchema):
    type: JSONSchemaInstanceType
    def __init__(
        self,
        schema,
        type,
        enum,
        const,
        format,
        title,
        description,
        anyOf,
        reference,
        definitions,
        default,
        deprecated,
        examples,
        properties,
        patternProperties,
        additionalProperties,
        propertyNames,
        prefixItems,
        items,
        contains,
        multipleOf,
        maximum,
        exclusiveMaximum,
        minimum,
        exclusiveMinimum,
        maxLength,
        minLength,
        pattern,
        maxItems,
        minItems,
        uniqueItems,
        maxContains,
        minContains,
        maxProperties,
        minProperties,
        required,
        dependentRequired,
    ) -> None: ...

class JSONArraySchema(JSONSchema):
    type: JSONSchemaInstanceType
    def __init__(
        self,
        schema,
        type,
        enum,
        const,
        format,
        title,
        description,
        anyOf,
        reference,
        definitions,
        default,
        deprecated,
        examples,
        properties,
        patternProperties,
        additionalProperties,
        propertyNames,
        prefixItems,
        items,
        contains,
        multipleOf,
        maximum,
        exclusiveMaximum,
        minimum,
        exclusiveMinimum,
        maxLength,
        minLength,
        pattern,
        maxItems,
        minItems,
        uniqueItems,
        maxContains,
        minContains,
        maxProperties,
        minProperties,
        required,
        dependentRequired,
    ) -> None: ...

class Context:
    dialect: JSONSchemaDialect
    definitions: Dict[str, JSONSchema]
    all_refs: Optional[bool]
    ref_prefix: Optional[str]
    def __init__(self, dialect, definitions, all_refs, ref_prefix) -> None: ...



import time
from typing import Callable

from dbt_common.events.types import RecordRetryException, RetryExternalCall
from dbt_common.exceptions import ConnectionError
from tarfile import ReadError
from gzip import BadGzipFile

import requests


def connection_exception_retry(fn: Callable, max_attempts: int, attempt: int = 0):
    """Handle connection retries gracefully.

    Attempts to run a function that makes an external call, if the call fails
    on a Requests exception or decompression issue (ReadError), it will be tried
    up to 5 more times.  All exceptions that Requests explicitly raises inherit from
    requests.exceptions.RequestException.  See https://github.com/dbt-labs/dbt-core/issues/4579
    for context on this decompression issues specifically.
    """
    try:
        return fn()
    except (
        requests.exceptions.RequestException,
        ReadError,
        EOFError,
        BadGzipFile,
    ) as exc:
        if attempt <= max_attempts - 1:
            # This import needs to be inline to avoid circular dependency
            from dbt_common.events.functions import fire_event

            fire_event(RecordRetryException(exc=str(exc)))
            fire_event(RetryExternalCall(attempt=attempt, max=max_attempts))
            time.sleep(1)
            return connection_exception_retry(fn, max_attempts, attempt + 1)
        else:
            raise ConnectionError("External connection exception occurred: " + str(exc))



from mashumaro.core.const import Sentinel
from mashumaro.dialect import Dialect
from mashumaro.types import Discriminator, SerializationStrategy
from typing import Any, Callable, Dict, List, Optional, Type, Union
from typing_extensions import Literal

TO_DICT_ADD_BY_ALIAS_FLAG: str
TO_DICT_ADD_OMIT_NONE_FLAG: str
ADD_DIALECT_SUPPORT: str
ADD_SERIALIZATION_CONTEXT: str
SerializationStrategyValueType = Union[SerializationStrategy, Dict[str, Union[str, Callable]]]

class BaseConfig:
    debug: bool
    code_generation_options: List[str]
    serialization_strategy: Dict[Any, SerializationStrategyValueType]
    aliases: Dict[str, str]
    serialize_by_alias: bool
    namedtuple_as_dict: bool
    allow_postponed_evaluation: bool
    dialect: Optional[Type[Dialect]]
    omit_none: Union[bool, Sentinel.MISSING]
    orjson_options: Optional[int]
    json_schema: Dict[str, Any]
    discriminator: Optional[Discriminator]
    lazy_compilation: bool



import abc
from _typeshed import Incomplete
from abc import ABC, abstractmethod
from mashumaro.core.meta.types.common import ValueSpec
from mashumaro.types import Discriminator
from typing import Optional, Tuple, Type

UnpackerRegistry: Incomplete

class AbstractUnpackerBuilder(ABC, metaclass=abc.ABCMeta):
    @abstractmethod
    def get_method_prefix(self) -> str: ...
    def build(self, spec: ValueSpec) -> str: ...

class UnionUnpackerBuilder(AbstractUnpackerBuilder):
    union_args: Incomplete
    def __init__(self, args: Tuple[Type, ...]) -> None: ...
    def get_method_prefix(self) -> str: ...

class TypeVarUnpackerBuilder(UnionUnpackerBuilder):
    def get_method_prefix(self) -> str: ...

class LiteralUnpackerBuilder(AbstractUnpackerBuilder):
    def get_method_prefix(self) -> str: ...

class DiscriminatedUnionUnpackerBuilder(AbstractUnpackerBuilder):
    discriminator: Incomplete
    base_variants: Incomplete
    def __init__(
        self, discriminator: Discriminator, base_variants: Optional[Tuple[Type, ...]] = ...
    ) -> None: ...
    def get_method_prefix(self) -> str: ...

class SubtypeUnpackerBuilder(DiscriminatedUnionUnpackerBuilder): ...



from typing import Any

from dbt_common.exceptions import CompilationError, DbtBaseException


class MacroReturn(DbtBaseException):
    """This is how we return a value from a macro, not an exception.

    Hack of all hacks
    """

    def __init__(self, value) -> None:
        self.value = value


class UndefinedMacroError(CompilationError):
    def __str__(self, prefix: str = "! ") -> str:
        msg = super().__str__(prefix)
        return (
            f"{msg}. This can happen when calling a macro that does "
            "not exist. Check for typos and/or install package dependencies "
            'with "dbt deps".'
        )


class UndefinedCompilationError(CompilationError):
    def __init__(self, name: str, node) -> None:
        self.name = name
        self.node = node
        self.msg = f"{self.name} is undefined"
        super().__init__(msg=self.msg)


class CaughtMacroError(CompilationError):
    def __init__(self, exc) -> None:
        self.exc = exc
        super().__init__(msg=str(exc))


class CaughtMacroErrorWithNodeError(CompilationError):
    def __init__(self, exc, node) -> None:
        self.exc = exc
        self.node = node
        super().__init__(msg=str(exc))


class JinjaRenderingError(CompilationError):
    pass


class MaterializationArgError(CompilationError):
    def __init__(self, name: str, argument: str) -> None:
        self.name = name
        self.argument = argument
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        msg = f"materialization '{self.name}' received unknown argument '{self.argument}'."
        return msg


class MacroNameNotStringError(CompilationError):
    def __init__(self, kwarg_value) -> None:
        self.kwarg_value = kwarg_value
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        msg = (
            f"The macro_name parameter ({self.kwarg_value}) "
            "to adapter.dispatch was not a string"
        )
        return msg


class MacrosSourcesUnWriteableError(CompilationError):
    def __init__(self, node) -> None:
        self.node = node
        msg = 'cannot "write" macros or sources'
        super().__init__(msg=msg)


class MacroArgTypeError(CompilationError):
    def __init__(self, method_name: str, arg_name: str, got_value: Any, expected_type) -> None:
        self.method_name = method_name
        self.arg_name = arg_name
        self.got_value = got_value
        self.expected_type = expected_type
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        got_type = type(self.got_value)
        msg = (
            f"'adapter.{self.method_name}' expects argument "
            f"'{self.arg_name}' to be of type '{self.expected_type}', instead got "
            f"{self.got_value} ({got_type})"
        )
        return msg


class MacroResultError(CompilationError):
    def __init__(self, freshness_macro_name: str, table):
        self.freshness_macro_name = freshness_macro_name
        self.table = table
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        msg = (
            f'Got an invalid result from "{self.freshness_macro_name}" '
            f"macro: {[tuple(r) for r in self.table]}"
        )

        return msg



import os
from contextvars import ContextVar, copy_context
from typing import List, Mapping, Optional, Iterator, Set

from dbt_common.constants import PRIVATE_ENV_PREFIX, SECRET_ENV_PREFIX
from dbt_common.record import Recorder


class CaseInsensitiveMapping(Mapping[str, str]):
    def __init__(self, env: Mapping[str, str]):
        self._env = {k.casefold(): (k, v) for k, v in env.items()}

    def __getitem__(self, key: str) -> str:
        return self._env[key.casefold()][1]

    def __len__(self) -> int:
        return len(self._env)

    def __iter__(self) -> Iterator[str]:
        for item in self._env.items():
            yield item[0]


class InvocationContext:
    def __init__(self, env: Mapping[str, str]):
        self._env: Mapping[str, str]

        env_public = {}
        env_private = {}

        for k, v in env.items():
            if k.startswith(PRIVATE_ENV_PREFIX):
                env_private[k] = v
            else:
                env_public[k] = v

        if os.name == "nt":
            self._env = CaseInsensitiveMapping(env_public)
        else:
            self._env = env_public

        self.name = "unset"
        self._env_secrets: Optional[List[str]] = None
        self._env_private = env_private
        self.recorder: Optional[Recorder] = None
        self._adapter_types: Set[str] = set()

        # If set to True later, this flag will prevent dbt from creating a new
        # invocation context for every invocation, which is useful for testing
        # scenarios.
        self.do_not_reset = False

        # This class will also eventually manage the invocation_id, flags, event manager, etc.

    @property
    def env(self) -> Mapping[str, str]:
        return self._env

    @property
    def env_private(self) -> Mapping[str, str]:
        return self._env_private

    @property
    def env_secrets(self) -> List[str]:
        if self._env_secrets is None:
            self._env_secrets = [
                v for k, v in self.env.items() if k.startswith(SECRET_ENV_PREFIX) and v.strip()
            ]
        return self._env_secrets

    @property
    def adapter_types(self) -> Set[str]:
        return self._adapter_types

    @adapter_types.setter
    def adapter_types(self, adapters: Set[str]) -> None:
        self._adapter_types = adapters

    def uses_adapter(self, adapter_type: str) -> None:
        self._adapter_types.add(adapter_type)


_INVOCATION_CONTEXT_VAR: ContextVar[InvocationContext] = ContextVar("DBT_INVOCATION_CONTEXT_VAR")


def reliably_get_invocation_var() -> ContextVar[InvocationContext]:
    invocation_var: Optional[ContextVar[InvocationContext]] = next(
        (cv for cv in copy_context() if cv.name == _INVOCATION_CONTEXT_VAR.name), None
    )

    if invocation_var is None:
        invocation_var = _INVOCATION_CONTEXT_VAR

    return invocation_var


def set_invocation_context(env: Mapping[str, str]) -> None:
    invocation_var = reliably_get_invocation_var()
    invocation_var.set(InvocationContext(env))


def get_invocation_context() -> InvocationContext:
    invocation_var = reliably_get_invocation_var()
    ctx = invocation_var.get()
    return ctx


def try_get_invocation_context() -> Optional[InvocationContext]:
    try:
        return get_invocation_context()
    except Exception:
        return None



from _typeshed import Incomplete
from mashumaro.dialect import Dialect as Dialect
from mashumaro.helper import pass_through as pass_through
from mashumaro.mixins.dict import DataClassDictMixin as DataClassDictMixin
from typing import Any, Callable, Dict, TypeVar, Type

T = TypeVar("T", bound="DataClassMessagePackMixin")
EncodedData = bytes
Encoder = Callable[[Any], EncodedData]
Decoder = Callable[[EncodedData], Dict[Any, Any]]

class MessagePackDialect(Dialect):
    serialization_strategy: Incomplete

def default_encoder(data: Any) -> EncodedData: ...
def default_decoder(data: EncodedData) -> Dict[Any, Any]: ...

class DataClassMessagePackMixin(DataClassDictMixin):
    def to_msgpack(self, encoder: Encoder = ..., **to_dict_kwargs: Any) -> EncodedData: ...
    @classmethod
    def from_msgpack(
        cls: Type[T], data: EncodedData, decoder: Decoder = ..., **from_dict_kwargs: Any
    ) -> T: ...



from typing import Any
from dbt_common.exceptions import CompilationError


# this is part of the context and also raised in dbt.contracts.relation.py
class DataclassNotDictError(CompilationError):
    def __init__(self, obj: Any):
        self.obj = obj
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        msg = (
            f'The object ("{self.obj}") was used as a dictionary. This '
            "capability has been removed from objects of this type."
        )

        return msg



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dbt_common.clients._jinja_blocks import Tag, TagIterator

from dbt_common.exceptions import CompilationError


class BlockDefinitionNotAtTopError(CompilationError):
    def __init__(self, tag_parser: "TagIterator", tag_start: int) -> None:
        self.tag_parser = tag_parser
        self.tag_start = tag_start
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        position = self.tag_parser.linepos(self.tag_start)
        msg = (
            f"Got a block definition inside control flow at {position}. "
            "All dbt block definitions must be at the top level"
        )
        return msg


class MissingCloseTagError(CompilationError):
    def __init__(self, block_type_name: str, linecount: int) -> None:
        self.block_type_name = block_type_name
        self.linecount = linecount
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        msg = (
            "Reached EOF without finding a close tag for "
            f"{self.block_type_name} (searched from line {self.linecount})"
        )
        return msg


class MissingControlFlowStartTagError(CompilationError):
    def __init__(self, tag: "Tag", expected_tag: str, tag_parser: "TagIterator") -> None:
        self.tag = tag
        self.expected_tag = expected_tag
        self.tag_parser = tag_parser
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        linepos = self.tag_parser.linepos(self.tag.start)
        msg = (
            f"Got an unexpected control flow end tag, got {self.tag.block_type_name} but "
            f"expected {self.expected_tag} next (@ {linepos})"
        )
        return msg


class NestedTagsError(CompilationError):
    def __init__(self, outer: "Tag", inner: "Tag") -> None:
        self.outer = outer
        self.inner = inner
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        msg = (
            f"Got nested tags: {self.outer.block_type_name} (started at {self.outer.start}) did "
            f"not have a matching {{{{% end{self.outer.block_type_name} %}}}} before a "
            f"subsequent {self.inner.block_type_name} was found (started at {self.inner.start})"
        )
        return msg


class UnexpectedControlFlowEndTagError(CompilationError):
    def __init__(self, tag: "Tag", expected_tag: str, tag_parser: "TagIterator") -> None:
        self.tag = tag
        self.expected_tag = expected_tag
        self.tag_parser = tag_parser
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        linepos = self.tag_parser.linepos(self.tag.start)
        msg = (
            f"Got an unexpected control flow end tag, got {self.tag.block_type_name} but "
            f"never saw a preceeding {self.expected_tag} (@ {linepos})"
        )
        return msg


class UnexpectedMacroEOFError(CompilationError):
    def __init__(self, expected_name: str, actual_name: str) -> None:
        self.expected_name = expected_name
        self.actual_name = actual_name
        super().__init__(msg=self.get_message())

    def get_message(self) -> str:
        msg = f'unexpected EOF, expected {self.expected_name}, got "{self.actual_name}"'
        return msg



# Since dbt-rpc does not do its own log setup, and since some events can
# currently fire before logs can be configured by setup_event_logger(), we
# create a default configuration with default settings and no file output.
from dbt_common.events.base_types import TCallback
from dbt_common.events.event_manager import IEventManager, EventManager

_EVENT_MANAGER: IEventManager = EventManager()


def get_event_manager() -> IEventManager:
    return _EVENT_MANAGER


def add_logger_to_manager(logger) -> None:
    _EVENT_MANAGER.add_logger(logger)


def add_callback_to_manager(callback: TCallback) -> None:
    _EVENT_MANAGER.add_callback(callback)


def ctx_set_event_manager(event_manager: IEventManager) -> None:
    global _EVENT_MANAGER
    _EVENT_MANAGER = event_manager


def cleanup_event_logger() -> None:
    # Reset to a no-op manager to release streams associated with logs. This is
    # especially important for tests, since pytest replaces the stdout stream
    # during test runs, and closes the stream after the test is over.
    _EVENT_MANAGER.loggers.clear()
    _EVENT_MANAGER.callbacks.clear()



from typing import Optional, Any

class Fore:
    RED: str = ...
    GREEN: str = ...
    YELLOW: str = ...

class Style:
    RESET_ALL: str = ...

def init(
    autoreset: bool = ...,
    convert: Optional[Any] = ...,
    strip: Optional[Any] = ...,
    wrap: bool = ...,
) -> None: ...



import copy
import datetime
from typing import Dict, Optional, TypeVar, Callable, Any, Tuple, Union, Type

from dbt_common.exceptions import DbtConfigError, RecursionError

K_T = TypeVar("K_T")
V_T = TypeVar("V_T")


def filter_null_values(input: Dict[K_T, Optional[V_T]]) -> Dict[K_T, V_T]:
    return {k: v for k, v in input.items() if v is not None}


class AttrDict(dict):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.__dict__ = self


def merge(*args):
    if len(args) == 0:
        return None

    if len(args) == 1:
        return args[0]

    lst = list(args)
    last = lst.pop(len(lst) - 1)

    return _merge(merge(*lst), last)


def _merge(a, b):
    to_return = a.copy()
    to_return.update(b)
    return to_return


def deep_merge(*args):
    """Deep merge dictionaries.

    Example:
    >>> dbt_common.utils.deep_merge(
    ...     {"a": 1, "b": 2, "c": 3}, {"a": 2}, {"a": 3, "b": 1}
    ... )  # noqa
    {'a': 3, 'b': 1, 'c': 3}
    From: http://stackoverflow.com/questions/20656135/python-deep-merge-dictionary-data
    """
    if len(args) == 0:
        return None

    if len(args) == 1:
        return copy.deepcopy(args[0])

    lst = list(args)
    last = copy.deepcopy(lst.pop(len(lst) - 1))

    return _deep_merge(deep_merge(*lst), last)


def _deep_merge(destination, source):
    if isinstance(source, dict):
        for key, value in source.items():
            deep_merge_item(destination, key, value)
        return destination


def deep_merge_item(destination, key, value):
    if isinstance(value, dict):
        node = destination.setdefault(key, {})
        destination[key] = deep_merge(node, value)
    elif isinstance(value, tuple) or isinstance(value, list):
        if key in destination:
            destination[key] = list(value) + list(destination[key])
        else:
            destination[key] = value
    else:
        destination[key] = value


def _deep_map_render(
    func: Callable[[Any, Tuple[Union[str, int], ...]], Any],
    value: Any,
    keypath: Tuple[Union[str, int], ...],
) -> Any:
    atomic_types: Tuple[Type[Any], ...] = (int, float, str, type(None), bool, datetime.date)

    ret: Any

    if isinstance(value, list):
        ret = [_deep_map_render(func, v, (keypath + (idx,))) for idx, v in enumerate(value)]
    elif isinstance(value, dict):
        ret = {k: _deep_map_render(func, v, (keypath + (str(k),))) for k, v in value.items()}
    elif isinstance(value, atomic_types):
        ret = func(value, keypath)
    else:
        container_types: Tuple[Type[Any], ...] = (list, dict)
        ok_types = container_types + atomic_types
        raise DbtConfigError(
            "in _deep_map_render, expected one of {!r}, got {!r}".format(ok_types, type(value))
        )

    return ret


def deep_map_render(func: Callable[[Any, Tuple[Union[str, int], ...]], Any], value: Any) -> Any:
    """This function renders a nested dictionary derived from a yaml file.

    It is used to render dbt_project.yml, profiles.yml, and
    schema files.

    It maps the function func() onto each non-container value in 'value'
    recursively, returning a new value. As long as func does not manipulate
    the value, then deep_map_render will also not manipulate it.

    value should be a value returned by `yaml.safe_load` or `json.load` - the
    only expected types are list, dict, native python number, str, NoneType,
    and bool.

    func() will be called on numbers, strings, Nones, and booleans. Its first
    parameter will be the value, and the second will be its keypath, an
    iterable over the __getitem__ keys needed to get to it.

    :raises: If there are cycles in the value, raises a
        dbt_common.exceptions.RecursionError
    """
    try:
        return _deep_map_render(func, value, ())
    except RuntimeError as exc:
        if "maximum recursion depth exceeded" in str(exc):
            raise RecursionError("Cycle detected in deep_map_render")
        raise



from dbt_common.exceptions.base import *  # noqa
from dbt_common.exceptions.events import *  # noqa
from dbt_common.exceptions.macros import *  # noqa
from dbt_common.exceptions.contracts import *  # noqa
from dbt_common.exceptions.connection import *  # noqa
from dbt_common.exceptions.system import *  # noqa
from dbt_common.exceptions.jinja import *  # noqa



import types
import typing
from _typeshed import Incomplete
from collections.abc import Generator
from dataclasses import Field
from mashumaro.config import (
    ADD_DIALECT_SUPPORT as ADD_DIALECT_SUPPORT,
    ADD_SERIALIZATION_CONTEXT as ADD_SERIALIZATION_CONTEXT,
    BaseConfig as BaseConfig,
    SerializationStrategyValueType as SerializationStrategyValueType,
    TO_DICT_ADD_BY_ALIAS_FLAG as TO_DICT_ADD_BY_ALIAS_FLAG,
    TO_DICT_ADD_OMIT_NONE_FLAG as TO_DICT_ADD_OMIT_NONE_FLAG,
)
from mashumaro.core.const import Sentinel as Sentinel
from mashumaro.core.helpers import ConfigValue as ConfigValue
from mashumaro.core.meta.code.lines import CodeLines as CodeLines
from mashumaro.core.meta.helpers import (
    get_args as get_args,
    get_class_that_defines_field as get_class_that_defines_field,
    get_class_that_defines_method as get_class_that_defines_method,
    get_literal_values as get_literal_values,
    get_name_error_name as get_name_error_name,
    hash_type_args as hash_type_args,
    is_class_var as is_class_var,
    is_dataclass_dict_mixin as is_dataclass_dict_mixin,
    is_dialect_subclass as is_dialect_subclass,
    is_init_var as is_init_var,
    is_literal as is_literal,
    is_optional as is_optional,
    is_type_var_any as is_type_var_any,
    resolve_type_params as resolve_type_params,
    substitute_type_params as substitute_type_params,
    type_name as type_name,
)
from mashumaro.core.meta.types.common import FieldContext as FieldContext, ValueSpec as ValueSpec
from mashumaro.core.meta.types.pack import PackerRegistry as PackerRegistry
from mashumaro.core.meta.types.unpack import (
    SubtypeUnpackerBuilder as SubtypeUnpackerBuilder,
    UnpackerRegistry as UnpackerRegistry,
)
from mashumaro.dialect import Dialect as Dialect
from mashumaro.exceptions import (
    BadDialect as BadDialect,
    BadHookSignature as BadHookSignature,
    InvalidFieldValue as InvalidFieldValue,
    MissingDiscriminatorError as MissingDiscriminatorError,
    MissingField as MissingField,
    SuitableVariantNotFoundError as SuitableVariantNotFoundError,
    ThirdPartyModuleNotFoundError as ThirdPartyModuleNotFoundError,
    UnresolvedTypeReferenceError as UnresolvedTypeReferenceError,
    UnserializableDataError as UnserializableDataError,
    UnserializableField as UnserializableField,
    UnsupportedDeserializationEngine as UnsupportedDeserializationEngine,
    UnsupportedSerializationEngine as UnsupportedSerializationEngine,
)
from mashumaro.types import Discriminator as Discriminator

__PRE_SERIALIZE__: str
__PRE_DESERIALIZE__: str
__POST_SERIALIZE__: str
__POST_DESERIALIZE__: str

class CodeBuilder:
    cls: Incomplete
    lines: Incomplete
    globals: Incomplete
    resolved_type_params: Incomplete
    field_classes: Incomplete
    initial_type_args: Incomplete
    dialect: Incomplete
    default_dialect: Incomplete
    allow_postponed_evaluation: Incomplete
    format_name: Incomplete
    decoder: Incomplete
    encoder: Incomplete
    encoder_kwargs: Incomplete
    def __init__(
        self,
        cls: typing.Type,
        type_args: typing.Tuple[typing.Type, ...] = ...,
        dialect: typing.Optional[typing.Type[Dialect]] = ...,
        first_method: str = ...,
        allow_postponed_evaluation: bool = ...,
        format_name: str = ...,
        decoder: typing.Optional[typing.Any] = ...,
        encoder: typing.Optional[typing.Any] = ...,
        encoder_kwargs: typing.Optional[typing.Dict[str, typing.Any]] = ...,
        default_dialect: typing.Optional[typing.Type[Dialect]] = ...,
    ) -> None: ...
    def reset(self) -> None: ...
    @property
    def namespace(self) -> typing.Mapping[typing.Any, typing.Any]: ...
    @property
    def annotations(self) -> typing.Dict[str, typing.Any]: ...
    def get_field_resolved_type_params(
        self, field_name: str
    ) -> typing.Dict[typing.Type, typing.Type]: ...
    def get_field_types(self, include_extras: bool = ...) -> typing.Dict[str, typing.Any]: ...
    @property
    def dataclass_fields(self) -> typing.Dict[str, Field]: ...
    @property
    def metadatas(self) -> typing.Dict[str, typing.Mapping[str, typing.Any]]: ...
    def get_field_default(self, name: str) -> typing.Any: ...
    def add_type_modules(self, *types_: typing.Type) -> None: ...
    def ensure_module_imported(self, module: types.ModuleType) -> None: ...
    def ensure_object_imported(
        self, obj: typing.Any, name: typing.Optional[str] = ...
    ) -> None: ...
    def add_line(self, line: str) -> None: ...
    def indent(self, expr: typing.Optional[str] = ...) -> typing.Generator[None, None, None]: ...
    def compile(self) -> None: ...
    def get_declared_hook(self, method_name: str) -> typing.Any: ...
    def add_unpack_method(self) -> None: ...
    def get_config(self, cls: Incomplete | None = ..., look_in_parents: bool = ...): ...
    def get_discriminator(self) -> typing.Optional[Discriminator]: ...
    def get_pack_method_flags(
        self, cls: typing.Optional[typing.Type] = ..., pass_encoder: bool = ...
    ) -> str: ...
    def get_unpack_method_flags(
        self, cls: typing.Optional[typing.Type] = ..., pass_decoder: bool = ...
    ) -> str: ...
    def get_pack_method_default_flag_values(
        self, cls: typing.Optional[typing.Type] = ..., pass_encoder: bool = ...
    ) -> str: ...
    def get_unpack_method_default_flag_values(self, pass_decoder: bool = ...) -> str: ...
    def is_code_generation_option_enabled(
        self, option: str, cls: typing.Optional[typing.Type] = ...
    ) -> bool: ...
    @classmethod
    def get_unpack_method_name(
        cls,
        type_args: typing.Iterable = ...,
        format_name: str = ...,
        decoder: typing.Optional[typing.Any] = ...,
    ) -> str: ...
    @classmethod
    def get_pack_method_name(
        cls,
        type_args: typing.Tuple[typing.Type, ...] = ...,
        format_name: str = ...,
        encoder: typing.Optional[typing.Any] = ...,
    ) -> str: ...
    def add_pack_method(self) -> None: ...
    def iter_serialization_strategies(
        self, metadata, ftype
    ) -> Generator[Incomplete, None, None]: ...



from typing import Any, Dict, Mapping, Type, TypeVar, Optional

T = TypeVar("T", bound="DataClassDictMixin")

class DataClassDictMixin:
    def __init_subclass__(cls: Type[T], **kwargs: Any) -> None: ...
    def to_dict(self, **kwargs: Any) -> dict: ...
    @classmethod
    def from_dict(cls, d: Mapping, **kwargs: Any) -> Any: ...
    @classmethod
    def __pre_deserialize__(cls: Type[T], d: Dict[Any, Any]) -> Dict[Any, Any]: ...
    @classmethod
    def __post_deserialize__(cls: Type[T], obj: T) -> T: ...
    def __pre_serialize__(self: T, context: Optional[Dict]) -> T: ...
    def __post_serialize__(self, d: Dict[Any, Any], context: Optional[Dict]) -> Dict[Any, Any]: ...



from typing import Optional

from dbt_common.exceptions import DbtInternalError


MACRO_PREFIX = "dbt_macro__"
DOCS_PREFIX = "dbt_docs__"


def get_dbt_macro_name(name: str) -> str:
    if name is None:
        raise DbtInternalError("Got None for a macro name!")
    return f"{MACRO_PREFIX}{name}"


def get_dbt_docs_name(name: str) -> str:
    if name is None:
        raise DbtInternalError("Got None for a doc name!")
    return f"{DOCS_PREFIX}{name}"


def get_materialization_macro_name(
    materialization_name: str, adapter_type: Optional[str] = None, with_prefix: bool = True
) -> str:
    if adapter_type is None:
        adapter_type = "default"
    name = f"materialization_{materialization_name}_{adapter_type}"
    return get_dbt_macro_name(name) if with_prefix else name


def get_docs_macro_name(docs_name: str, with_prefix: bool = True) -> str:
    return get_dbt_docs_name(docs_name) if with_prefix else docs_name


def get_test_macro_name(test_name: str, with_prefix: bool = True) -> str:
    name = f"test_{test_name}"
    return get_dbt_macro_name(name) if with_prefix else name



from mashumaro.exceptions import MissingField as MissingField
from mashumaro.helper import field_options as field_options, pass_through as pass_through
from mashumaro.mixins.dict import DataClassDictMixin as DataClassDictMixin



import concurrent.futures
from contextlib import contextmanager
from typing import Protocol, Optional

from dbt_common.context import (
    get_invocation_context,
    reliably_get_invocation_var,
    InvocationContext,
)


class ConnectingExecutor(concurrent.futures.Executor):
    def submit_connected(self, adapter, conn_name, func, *args, **kwargs):
        def connected(conn_name, func, *args, **kwargs):
            with self.connection_named(adapter, conn_name):
                return func(*args, **kwargs)

        return self.submit(connected, conn_name, func, *args, **kwargs)


# a little concurrent.futures.Executor for single-threaded mode
class SingleThreadedExecutor(ConnectingExecutor):
    def submit(*args, **kwargs):
        # this basic pattern comes from concurrent.futures.Executor itself,
        # but without handling the `fn=` form.
        if len(args) >= 2:
            self, fn, *args = args
        elif not args:
            raise TypeError(
                "descriptor 'submit' of 'SingleThreadedExecutor' object needs an argument"
            )
        else:
            raise TypeError(
                "submit expected at least 1 positional argument, got %d" % (len(args) - 1)
            )
        fut = concurrent.futures.Future()
        try:
            result = fn(*args, **kwargs)
        except Exception as exc:
            fut.set_exception(exc)
        else:
            fut.set_result(result)
        return fut

    @contextmanager
    def connection_named(self, adapter, name):
        yield


class MultiThreadedExecutor(
    ConnectingExecutor,
    concurrent.futures.ThreadPoolExecutor,
):
    @contextmanager
    def connection_named(self, adapter, name):
        with adapter.connection_named(name):
            yield


class ThreadedArgs(Protocol):
    single_threaded: bool


class HasThreadingConfig(Protocol):
    args: ThreadedArgs
    threads: Optional[int]


def _thread_initializer(invocation_context: InvocationContext) -> None:
    invocation_var = reliably_get_invocation_var()
    invocation_var.set(invocation_context)


def executor(config: HasThreadingConfig) -> ConnectingExecutor:
    if config.args.single_threaded:
        return SingleThreadedExecutor()
    else:
        return MultiThreadedExecutor(
            max_workers=config.threads,
            initializer=_thread_initializer,  # type: ignore
            initargs=(get_invocation_context(),),  # type: ignore
        )



from typing import Any, Optional

DEFAULT_NULL_VALUES: Any

class DataType:
    null_values: Any = ...
    def __init__(self, null_values: Any = ...) -> None: ...
    def test(self, d: Any): ...
    def cast(self, d: Any) -> None: ...
    def csvify(self, d: Any): ...
    def jsonify(self, d: Any): ...

DEFAULT_TRUE_VALUES: Any
DEFAULT_FALSE_VALUES: Any

class Boolean(DataType):
    true_values: Any = ...
    false_values: Any = ...
    def __init__(
        self, true_values: Any = ..., false_values: Any = ..., null_values: Any = ...
    ) -> None: ...
    def cast(self, d: Any): ...
    def jsonify(self, d: Any): ...

ZERO_DT: Any

class Date(DataType):
    date_format: Any = ...
    parser: Any = ...
    def __init__(self, date_format: Optional[Any] = ..., **kwargs: Any) -> None: ...
    def cast(self, d: Any): ...
    def csvify(self, d: Any): ...
    def jsonify(self, d: Any): ...

class DateTime(DataType):
    datetime_format: Any = ...
    timezone: Any = ...
    def __init__(
        self, datetime_format: Optional[Any] = ..., timezone: Optional[Any] = ..., **kwargs: Any
    ) -> None: ...
    def cast(self, d: Any): ...
    def csvify(self, d: Any): ...
    def jsonify(self, d: Any): ...

DEFAULT_CURRENCY_SYMBOLS: Any
POSITIVE: Any
NEGATIVE: Any

class Number(DataType):
    locale: Any = ...
    currency_symbols: Any = ...
    group_symbol: Any = ...
    decimal_symbol: Any = ...
    def __init__(
        self,
        locale: str = ...,
        group_symbol: Optional[Any] = ...,
        decimal_symbol: Optional[Any] = ...,
        currency_symbols: Any = ...,
        **kwargs: Any,
    ) -> None: ...
    def cast(self, d: Any): ...
    def jsonify(self, d: Any): ...

class TimeDelta(DataType):
    def cast(self, d: Any): ...

class Text(DataType):
    cast_nulls: Any = ...
    def __init__(self, cast_nulls: bool = ..., **kwargs: Any) -> None: ...
    def cast(self, d: Any): ...



# Prefix which identifies environment variables which contains secrets.
SECRET_ENV_PREFIX = "DBT_ENV_SECRET"

# Prefix which identifies environment variables that should not be visible
# via macros, flags, or other user-facing mechanisms.
PRIVATE_ENV_PREFIX = "DBT_ENV_PRIVATE"

# Prefix for dbt engine environment varaibles that are user settable
ENGINE_ENV_PREFIX = "DBT_ENGINE"



from mashumaro.dialect import Dialect
from typing import Any, Dict, Optional, Tuple, Type

def compile_mixin_packer(
    cls,
    format_name: str = ...,
    dialect: Optional[Type[Dialect]] = ...,
    encoder: Any = ...,
    encoder_kwargs: Optional[Dict[str, Dict[str, Tuple[str, Any]]]] = ...,
) -> None: ...
def compile_mixin_unpacker(
    cls, format_name: str = ..., dialect: Optional[Type[Dialect]] = ..., decoder: Any = ...
) -> None: ...



from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, List

from dbt_common.dataclass_schema import dbtClassMixin


class ConstraintType(str, Enum):
    check = "check"
    not_null = "not_null"
    unique = "unique"
    primary_key = "primary_key"
    foreign_key = "foreign_key"
    custom = "custom"

    @classmethod
    def is_valid(cls, item) -> bool:
        try:
            cls(item)
        except ValueError:
            return False
        return True


@dataclass
class ColumnLevelConstraint(dbtClassMixin):
    type: ConstraintType
    name: Optional[str] = None
    # expression is a user-provided field that will depend on the constraint type.
    # It could be a predicate (check type), or a sequence sql keywords (e.g. unique type),
    # so the vague naming of 'expression' is intended to capture this range.
    expression: Optional[str] = None
    warn_unenforced: bool = (
        True  # Warn if constraint cannot be enforced by platform but will be in DDL
    )
    warn_unsupported: bool = (
        True  # Warn if constraint is not supported by the platform and won't be in DDL
    )
    to: Optional[str] = None
    to_columns: List[str] = field(default_factory=list)


@dataclass
class ModelLevelConstraint(ColumnLevelConstraint):
    columns: List[str] = field(default_factory=list)



from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Type,
)

def get_type_origin(typ: Type[Any]) -> Type[Any]: ...
def get_generic_name(typ: Type[Any], short: bool = ...) -> str: ...
def get_args(typ: Optional[Type[Any]]) -> Tuple[Type[Any], ...]: ...
def get_literal_values(typ: Type[Any]) -> Tuple[Any, ...]: ...
def type_name(
    typ: Optional[Type[Any]],
    short: bool = ...,
    resolved_type_params: Optional[Dict[Type[Any], Type[Any]]] = ...,
    is_type_origin: bool = ...,
    none_type_as_none: bool = ...,
) -> str: ...
def is_special_typing_primitive(typ: Any) -> bool: ...
def is_generic(typ: Type[Any]) -> bool: ...
def is_typed_dict(typ: Type[Any]) -> bool: ...
def is_named_tuple(typ: Type[Any]) -> bool: ...
def is_new_type(typ: Type[Any]) -> bool: ...
def is_union(typ: Type[Any]) -> bool: ...
def is_optional(
    typ: Type[Any], resolved_type_params: Optional[Dict[Type[Any], Type[Any]]] = ...
) -> bool: ...
def is_annotated(typ: Type[Any]) -> bool: ...
def get_type_annotations(typ: Type[Any]) -> Sequence[Any]: ...
def is_literal(typ: Type[Any]) -> bool: ...
def not_none_type_arg(
    type_args: Tuple[Type[Any], ...],
    resolved_type_params: Optional[Dict[Type[Any], Type[Any]]] = ...,
) -> Optional[Type[Any]]: ...
def is_type_var(typ: Type[Any]) -> bool: ...
def is_type_var_any(typ: Type[Any]) -> bool: ...
def is_class_var(typ: Type[Any]) -> bool: ...
def is_final(typ: Type[Any]) -> bool: ...
def is_init_var(typ: Type[Any]) -> bool: ...
def get_class_that_defines_method(method_name: str, cls: Type[Any]) -> Optional[Type[Any]]: ...
def get_class_that_defines_field(field_name: str, cls: Type[Any]) -> Optional[Type[Any]]: ...
def is_dataclass_dict_mixin(typ: Type[Any]) -> bool: ...
def is_dataclass_dict_mixin_subclass(typ: Type[Any]) -> bool: ...
def collect_type_params(typ: Type[Any]) -> Sequence[Type[Any]]: ...
def resolve_type_params(
    typ: Type[Any], type_args: Sequence[Type[Any]] = ..., include_bases: bool = ...
) -> Dict[Type[Any], Dict[Type[Any], Type[Any]]]: ...
def substitute_type_params(
    typ: Type[Any], substitutions: Dict[Type[Any], Type[Any]]
) -> Type[Any]: ...
def get_name_error_name(e: NameError) -> str: ...
def is_dialect_subclass(typ: Type[Any]) -> bool: ...
def is_self(typ: Type[Any]) -> bool: ...
def is_required(typ: Type[Any]) -> bool: ...
def is_not_required(typ: Type[Any]) -> bool: ...
def get_function_arg_annotation(
    function: Callable[[Any], Any],
    arg_name: Optional[str] = ...,
    arg_pos: Optional[int] = ...,
) -> Type[Any]: ...
def get_function_return_annotation(function: Callable[[Any], Any]) -> Type[Any]: ...
def is_unpack(typ: Type[Any]) -> bool: ...
def is_type_var_tuple(typ: Type[Any]) -> bool: ...
def hash_type_args(type_args: Iterable[Type[Any]]) -> str: ...
def iter_all_subclasses(cls: Type[Any]) -> Iterator[Type[Any]]: ...



from dbt_common.dataclass_schema import StrEnum


class OnConfigurationChangeOption(StrEnum):
    Apply = "apply"
    Continue = "continue"
    Fail = "fail"

    @classmethod
    def default(cls) -> "OnConfigurationChangeOption":
        return cls.Apply



# This is useful for proto generated classes in particular, since
# the default for protobuf for strings is the empty string, so
# Optional[str] types don't work for generated Python classes.
from typing import Any, Dict, Mapping, Optional


def cast_to_str(string: Optional[str]) -> str:
    if string is None:
        return ""
    else:
        return string


def cast_to_int(integer: Optional[int]) -> int:
    if integer is None:
        return 0
    else:
        return integer


def cast_dict_to_dict_of_strings(dct: Mapping[Any, Any]) -> Dict[str, str]:
    new_dct: Dict[str, str] = {}

    for k, v in dct.items():
        new_dct[str(k)] = str(v)
    return new_dct



"""The record module provides a record/replay mechanism for recording dbt's
interactions with external systems during a command invocation, so that the
command can be re-run later with the recording 'replayed' to dbt.

The rationale for and architecture of this module are described in detail in the
docs/guides/record_replay.md document in this repository.
"""
import functools
import dataclasses
import inspect
import json
import os

from enum import Enum
from threading import Lock
from typing import Any, Callable, Dict, List, Mapping, Optional, TextIO, Tuple, Type

from mashumaro import field_options
from mashumaro.mixins.json import DataClassJSONMixin
from mashumaro.types import SerializationStrategy

import contextvars


RECORDED_BY_HIGHER_FUNCTION = contextvars.ContextVar("RECORDED_BY_HIGHER_FUNCTION", default=False)


class Record:
    """An instance of this abstract Record class represents a request made by dbt
    to an external process or the operating system. The 'params' are the arguments
    to the request, and the 'result' is what is returned."""

    params_cls: type
    result_cls: Optional[type] = None
    group: Optional[str] = None

    def __init__(self, params, result, seq=None) -> None:
        self.params = params
        self.result = result
        self.seq = seq

    def to_dict(self) -> Dict[str, Any]:
        return {
            "params": self.params._to_dict()
            if hasattr(self.params, "_to_dict")
            else dataclasses.asdict(self.params),
            "result": self.result._to_dict()
            if hasattr(self.result, "_to_dict")
            else dataclasses.asdict(self.result)
            if self.result is not None
            else None,
            "seq": self.seq,
        }

    @classmethod
    def from_dict(cls, dct: Mapping) -> "Record":
        p = (
            cls.params_cls._from_dict(dct["params"])
            if hasattr(cls.params_cls, "_from_dict")
            else cls.params_cls(**dct["params"])
        )
        r = (
            cls.result_cls._from_dict(dct["result"])  # type: ignore
            if hasattr(cls.result_cls, "_from_dict")
            else cls.result_cls(**dct["result"])
            if cls.result_cls is not None
            else None
        )
        s = dct.get("seq", None)
        return cls(params=p, result=r, seq=s)


class Diff:
    def __init__(self, current_recording_path: str, previous_recording_path: str) -> None:
        # deepdiff is expensive to import, so we only do it here when we need it
        from deepdiff import DeepDiff  # type: ignore

        self.diff = DeepDiff

        self.current_recording_path = current_recording_path
        self.previous_recording_path = previous_recording_path

    def diff_query_records(self, current: List, previous: List) -> Dict[str, Any]:
        # some of the table results are returned as a stringified list of dicts that don't
        # diff because order isn't consistent. convert it into a list of dicts so it can
        # be diffed ignoring order

        for i in range(len(current)):
            if current[i].get("result").get("table") is not None:
                current[i]["result"]["table"] = json.loads(current[i]["result"]["table"])
        for i in range(len(previous)):
            if previous[i].get("result").get("table") is not None:
                previous[i]["result"]["table"] = json.loads(previous[i]["result"]["table"])

        return self.diff(previous, current, ignore_order=True, verbose_level=2)

    def diff_env_records(self, current: List, previous: List) -> Dict[str, Any]:
        # The mode and filepath may change.  Ignore them.

        exclude_paths = [
            "root[0]['result']['env']['DBT_RECORDER_FILE_PATH']",
            "root[0]['result']['env']['DBT_RECORDER_MODE']",
        ]

        return self.diff(
            previous, current, ignore_order=True, verbose_level=2, exclude_paths=exclude_paths
        )

    def diff_default(self, current: List, previous: List) -> Dict[str, Any]:
        return self.diff(previous, current, ignore_order=True, verbose_level=2)

    def calculate_diff(self) -> Dict[str, Any]:
        with open(self.current_recording_path) as current_recording:
            current_dct = json.load(current_recording)

        with open(self.previous_recording_path) as previous_recording:
            previous_dct = json.load(previous_recording)

        diff = {}
        for record_type in current_dct:
            if record_type == "QueryRecord":
                diff[record_type] = self.diff_query_records(
                    current_dct[record_type], previous_dct[record_type]
                )
            elif record_type == "GetEnvRecord":
                diff[record_type] = self.diff_env_records(
                    current_dct[record_type], previous_dct[record_type]
                )
            else:
                diff[record_type] = self.diff_default(
                    current_dct[record_type], previous_dct[record_type]
                )

        return diff


class RecorderMode(Enum):
    RECORD = 1
    REPLAY = 2
    DIFF = 3  # records and does diffing


class Recorder:
    _record_cls_by_name: Dict[str, Type] = {}
    _record_name_by_params_name: Dict[str, str] = {}
    _auto_serialization_strategies: Dict[Type, SerializationStrategy] = {}

    def __init__(
        self,
        mode: RecorderMode,
        types: Optional[List],
        current_recording_path: str = "recording.json",
        previous_recording_path: Optional[str] = None,
        in_memory: bool = False,
    ) -> None:
        self.mode = mode
        self.recorded_types = types
        self._records_by_type: Dict[str, List[Record]] = {}
        self._unprocessed_records_by_type: Dict[str, List[Dict[str, Any]]] = {}
        self._replay_diffs: List["Diff"] = []
        self.diff: Optional[Diff] = None
        self.previous_recording_path = previous_recording_path
        self.current_recording_path = current_recording_path

        if self.previous_recording_path is not None and self.mode in (
            RecorderMode.REPLAY,
            RecorderMode.DIFF,
        ):
            self.diff = Diff(
                current_recording_path=self.current_recording_path,
                previous_recording_path=self.previous_recording_path,
            )

            if self.mode == RecorderMode.REPLAY:
                self._unprocessed_records_by_type = self.load(self.previous_recording_path)

        self._counter = 0
        self._counter_lock = Lock()

        self._record_added = False
        self._recording_file: Optional[TextIO] = None
        self._recording_file_lock = Lock()
        if mode == RecorderMode.RECORD and not in_memory:
            self._recording_file = open(current_recording_path, "w")
            self._recording_file.write("[")

    def __del__(self):
        self.clean_up_stream()

    @classmethod
    def register_record_type(cls, rec_type) -> Any:
        cls._record_cls_by_name[rec_type.__name__] = rec_type
        cls._record_name_by_params_name[rec_type.params_cls.__name__] = rec_type.__name__
        return rec_type

    def add_record(self, record: Record) -> None:
        rec_cls_name = record.__class__.__name__  # type: ignore

        with self._counter_lock:
            record.seq = self._counter
            self._counter += 1

        if self._recording_file is not None:
            # Lock recording file during streamed recording to avoid race conditions across recording threads
            with self._recording_file_lock:
                if self._record_added:
                    self._recording_file.write(",")
                try:
                    dct = Recorder._get_tagged_dict(record, rec_cls_name)
                    json.dump(dct, self._recording_file)
                    self._record_added = True
                except Exception as e:
                    json.dump(
                        {"type": "RecordingError", "record_type": rec_cls_name, "error": str(e)},
                        self._recording_file,
                    )
        else:
            if rec_cls_name not in self._records_by_type:
                self._records_by_type[rec_cls_name] = []
            self._records_by_type[rec_cls_name].append(record)

    def pop_matching_record(self, params: Any) -> Optional[Record]:
        rec_type_name = self._record_name_by_params_name.get(type(params).__name__)

        if rec_type_name is None:
            raise Exception(
                f"A record of type {type(params).__name__} was requested, but no such type has been registered."
            )

        self._ensure_records_processed(rec_type_name)
        records = self._records_by_type[rec_type_name]
        match: Optional[Record] = None
        for rec in records:
            if rec.params == params:
                match = rec
                records.remove(match)
                break

        return match

    def write_json(self, out_stream: TextIO):
        d = self._to_list()
        json.dump(d, out_stream)

    def write(self) -> None:
        if self._recording_file is not None:
            self.clean_up_stream()
        else:
            with open(self.current_recording_path, "w") as file:
                self.write_json(file)

    def clean_up_stream(self) -> None:
        if self._recording_file is not None:
            self._recording_file.write("]")
            self._recording_file.close()
            self._recording_file = None

    @staticmethod
    def _get_tagged_dict(record: Record, record_type: str) -> Dict:
        d = record.to_dict()
        d["type"] = record_type
        return d

    def _to_list(self) -> List[Dict]:
        record_list: List[Dict] = []
        for record_type in self._records_by_type:
            record_list.extend(
                Recorder._get_tagged_dict(r, record_type)
                for r in self._records_by_type[record_type]
            )

        record_list.sort(key=lambda r: r["seq"])

        return record_list

    @classmethod
    def load(cls, file_name: str) -> Dict[str, List[Dict[str, Any]]]:
        with open(file_name) as file:
            return cls.load_json(file)

    @classmethod
    def load_json(cls, in_stream: TextIO) -> Dict[str, List[Dict[str, Any]]]:
        return json.load(in_stream)

    def _ensure_records_processed(self, record_type_name: str) -> None:
        if record_type_name in self._records_by_type:
            return

        rec_list = []
        record_cls = self._record_cls_by_name[record_type_name]
        for record_dct in self._unprocessed_records_by_type[record_type_name]:
            rec = record_cls.from_dict(record_dct)
            rec_list.append(rec)  # type: ignore
        self._records_by_type[record_type_name] = rec_list

    def expect_record(self, params: Any) -> Any:
        record = self.pop_matching_record(params)

        if record is None:
            raise Exception()

        if record.result is None:
            return None

        result_tuple = dataclasses.astuple(record.result)
        return result_tuple[0] if len(result_tuple) == 1 else result_tuple

    def write_diffs(self, diff_file_name) -> None:
        assert self.diff is not None
        with open(diff_file_name, "w") as f:
            json.dump(self.diff.calculate_diff(), f)

    def print_diffs(self) -> None:
        assert self.diff is not None
        print(repr(self.diff.calculate_diff()))

    @classmethod
    def register_serialization_strategy(
        cls, t: Type, serialization_strategy: SerializationStrategy
    ) -> None:
        cls._auto_serialization_strategies[t] = serialization_strategy


def get_record_mode_from_env() -> Optional[RecorderMode]:
    """
    Get the record mode from the environment variables.

    If the mode is not set to 'RECORD', 'DIFF' or 'REPLAY', return None.
    Expected format: 'DBT_RECORDER_MODE=RECORD'
    """
    record_mode = os.environ.get("DBT_RECORDER_MODE")

    if record_mode is None:
        return None

    if record_mode.lower() == "record":
        return RecorderMode.RECORD
    # diffing requires a file path, otherwise treat as noop
    elif record_mode.lower() == "diff" and os.environ.get("DBT_RECORDER_FILE_PATH") is not None:
        return RecorderMode.DIFF
    # replaying requires a file path, otherwise treat as noop
    elif record_mode.lower() == "replay" and os.environ.get("DBT_RECORDER_FILE_PATH") is not None:
        return RecorderMode.REPLAY

    # if you don't specify record/replay it's a noop
    return None


def get_record_types_from_env() -> Optional[List]:
    """
    Get the record subset from the environment variables.

    If no types are provided, there will be no filtering.
    Invalid types will be ignored.
    Expected format: 'DBT_RECORDER_TYPES=Database,FileLoadRecord'
    """
    record_types_str = os.environ.get("DBT_RECORDER_TYPES")

    # if all is specified we don't want any type filtering
    if record_types_str is None or record_types_str.lower == "all":
        return None

    return record_types_str.split(",")


def get_record_types_from_dict(fp: str) -> List:
    """Get the record subset from the dict."""
    with open(fp) as file:
        loaded_dct = json.load(file)
    return list(loaded_dct.keys())


def auto_record_function(
    record_name: str,
    method: bool = True,
    group: Optional[str] = None,
    index_on_thread_name: bool = True,
) -> Callable:
    """This is the @auto_record_function decorator. It works in a similar way to
    the @record_function decorator, except automatically generates boilerplate
    classes for the Record, Params, and Result classes which would otherwise be
    needed. That makes it suitable for quickly adding record support to simple
    functions with simple parameters."""
    return functools.partial(
        _record_function_inner,
        record_name,
        method,
        False,
        None,
        group,
        index_on_thread_name,
    )


def record_function(
    record_type,
    method: bool = False,
    tuple_result: bool = False,
    id_field_name: Optional[str] = None,
    index_on_thread_id: bool = False,
) -> Callable:
    """This is the @record_function decorator, which marks functions which will
    have their function calls recorded during record mode, and mocked out with
    previously recorded replay data during replay."""
    return functools.partial(
        _record_function_inner,
        record_type,
        method,
        tuple_result,
        id_field_name,
        None,
        index_on_thread_id,
    )


def _get_arg_fields(
    spec: inspect.FullArgSpec,
    skip_first: bool = False,
) -> List[Tuple[str, Optional[Type], dataclasses.Field]]:
    arg_fields = []
    defaults = len(spec.defaults) if spec.defaults else 0
    for i, arg_name in enumerate(spec.args):
        if skip_first and i == 0:
            continue
        annotation = spec.annotations.get(arg_name)
        if annotation is None:
            raise Exception("Recorded functions must have type annotations.")
        field = _get_field(arg_name, annotation)
        if i >= len(spec.args) - defaults:
            field[2].default = (
                spec.defaults[i - len(spec.args) + defaults] if spec.defaults else None
            )
        arg_fields.append(field)
    return arg_fields


def _get_field(field_name: str, t: Type) -> Tuple[str, Optional[Type], dataclasses.Field]:
    dc_field: dataclasses.Field = dataclasses.field()
    strat = Recorder._auto_serialization_strategies.get(t)
    if strat is not None:
        dc_field.metadata = field_options(serialization_strategy=Recorder._auto_serialization_strategies[t])  # type: ignore

    return field_name, t, dc_field


@dataclasses.dataclass
class AutoValues(DataClassJSONMixin):
    def _to_dict(self):
        return self.to_dict()

    @classmethod
    def _from_dict(cls, data):
        return cls.from_dict(data)


def _record_function_inner(
    record_type,
    method,
    tuple_result,
    id_field_name,
    group,
    index_on_thread_id,
    func_to_record,
):
    recorded_types = get_record_types_from_env()
    if recorded_types is not None and not (
        getattr(record_type, "__name__", record_type) in recorded_types
        or getattr(record_type, "group", group) in recorded_types
    ):
        return func_to_record

    if isinstance(record_type, str):
        return_type = inspect.signature(func_to_record).return_annotation
        fields = _get_arg_fields(inspect.getfullargspec(func_to_record), method)
        if index_on_thread_id:
            id_field_name = "thread_id"
            fields.insert(0, _get_field("thread_id", str))
        params_cls = dataclasses.make_dataclass(
            f"{record_type}Params", fields, bases=(AutoValues,)
        )
        result_cls = (
            None
            if return_type is None or return_type == inspect._empty
            else dataclasses.make_dataclass(
                f"{record_type}Result",
                [_get_field("return_val", return_type)],
                bases=(AutoValues,),
            )
        )

        record_type = type(
            f"{record_type}Record",
            (Record,),
            {"params_cls": params_cls, "result_cls": result_cls, "group": group},
        )

        Recorder.register_record_type(record_type)

    @functools.wraps(func_to_record)
    def record_replay_wrapper(*args, **kwargs) -> Any:
        recorder: Optional[Recorder] = None
        try:
            from dbt_common.context import get_invocation_context

            recorder = get_invocation_context().recorder
        except LookupError:
            pass

        call_args = args

        if recorder is None:
            return func_to_record(*call_args, **kwargs)

        if recorder.recorded_types is not None and not (
            record_type.__name__ in recorder.recorded_types
            or record_type.group in recorder.recorded_types
        ):
            return func_to_record(*call_args, **kwargs)

        # For methods, peel off the 'self' argument before calling the
        # params constructor.
        param_args = args[1:] if method else args
        if method and id_field_name is not None:
            if index_on_thread_id:
                from dbt_common.events.contextvars import get_node_info

                node_info = get_node_info()
                if node_info and "unique_id" in node_info:
                    thread_name = node_info["unique_id"]
                else:
                    from dbt_common.context import get_invocation_context

                    thread_name = get_invocation_context().name
                param_args = (thread_name,) + param_args
            else:
                param_args = (getattr(args[0], id_field_name),) + param_args

        # Build params - this can be dangerous if a subclass overrides the method in such a way that
        # changes the signature of the base recorded method, and so is wrapped in a try/except.
        params = None
        try:
            try:
                # Omits any additional properties that are not fields of the params class
                params_dict = {
                    field.name: value
                    for field, value in zip(dataclasses.fields(record_type.params_cls), param_args)
                }
                params_dict.update(kwargs)
                params = record_type.params_cls._from_dict(params_dict)
            except Exception:
                params = record_type.params_cls(*param_args, **kwargs)
        except Exception:
            # Unfortunately it is not possible to fire an event here because it would cause a circular import
            # This means we lose visibility into issues using record_type.params_cls(...), but it is better than crashing the entire node or command
            pass

        include = True
        if params is not None and hasattr(params, "_include"):
            include = params._include()

        if not include:
            return func_to_record(*call_args, **kwargs)

        if recorder.mode == RecorderMode.REPLAY and params is not None:
            return recorder.expect_record(params)
        if RECORDED_BY_HIGHER_FUNCTION.get():
            return func_to_record(*call_args, **kwargs)

        RECORDED_BY_HIGHER_FUNCTION.set(True)
        r = func_to_record(*call_args, **kwargs)
        result = None

        # Gracefully handle the case where the result is not serializable
        try:
            result = (
                None
                if record_type.result_cls is None
                else record_type.result_cls(*r)
                if tuple_result
                else record_type.result_cls(r)
            )
        except Exception:
            pass

        RECORDED_BY_HIGHER_FUNCTION.set(False)
        if params is not None:
            recorder.add_record(record_type(params=params, result=result))
        return r

    setattr(
        record_replay_wrapper,
        "_record_metadata",
        {
            "record_type": record_type,
            "method": method,
            "tuple_result": tuple_result,
            "id_field_name": id_field_name,
            "group": group,
            "index_on_thread_id": index_on_thread_id,
        },
    )

    return record_replay_wrapper


def _is_classmethod(method):
    b = inspect.ismethod(method) and isinstance(method.__self__, type)
    return b


def supports_replay(cls):
    """Class decorator which adds record/replay support for a class. In particular,
    this decorator ensures that calls to overriden functions are still recorded."""

    # When record/replay is inactive, do nothing.
    if get_record_mode_from_env() is None:
        return cls

    # Replace the __init_subclass__ method of this class so that when it
    # is subclassed, methods on the new subclass which override recorded
    # functions are modified to be recorded as well.
    original_init_subclass = cls.__init_subclass__

    @classmethod
    def wrapping_init_subclass(sub_cls):
        for method_name in dir(cls):
            method = getattr(cls, method_name)
            metadata = getattr(method, "_record_metadata", None)
            if method and getattr(method, "_record_metadata", None):
                sub_method = getattr(sub_cls, method_name, None)
                sub_method_metadata = getattr(sub_method, "_record_metadata", None)

                # Handle classmethod overrides. This logic goes above and beyond
                # to handle the situation where the method is a classmethod, but
                # the submethod is not (and therefore lacks a __func__ attribute).
                override_as_classmethod = _is_classmethod(method) and hasattr(
                    sub_method, "__func__"
                )

                if not sub_method_metadata:
                    recorded_sub_method = _record_function_inner(
                        metadata["record_type"],
                        metadata["method"],
                        metadata["tuple_result"],
                        metadata["id_field_name"],
                        metadata["group"],
                        metadata["index_on_thread_id"],
                        sub_method.__func__
                        if override_as_classmethod
                        else sub_method,  # Unwrap if method and submethod are both classmethods
                    )

                    if _is_classmethod(method) and hasattr(sub_method, "__func__"):
                        # Rewrap if method and submethod are both classmethods
                        recorded_sub_method = classmethod(recorded_sub_method)

                    setattr(
                        sub_cls,
                        method_name,
                        recorded_sub_method,
                    )

        original_init_subclass()

    cls.__init_subclass__ = wrapping_init_subclass

    return cls



from collections.abc import Sequence

from typing import Any, Optional, Callable, Iterable, Dict, Union

from . import data_types as data_types
from .data_types import (
    Text as Text,
    Number as Number,
    Boolean as Boolean,
    DateTime as DateTime,
    Date as Date,
    TimeDelta as TimeDelta,
)

class MappedSequence(Sequence):
    def __init__(self, values: Any, keys: Optional[Any] = ...) -> None: ...
    def __unicode__(self): ...
    def __getitem__(self, key: Any): ...
    def __setitem__(self, key: Any, value: Any) -> None: ...
    def __iter__(self): ...
    def __len__(self): ...
    def __eq__(self, other: Any): ...
    def __ne__(self, other: Any): ...
    def __contains__(self, value: Any): ...
    def keys(self): ...
    def values(self): ...
    def items(self): ...
    def get(self, key: Any, default: Optional[Any] = ...): ...
    def dict(self): ...

class Row(MappedSequence): ...

class Table:
    def __init__(
        self,
        rows: Any,
        column_names: Optional[Any] = ...,
        column_types: Optional[Any] = ...,
        row_names: Optional[Any] = ...,
        _is_fork: bool = ...,
    ) -> None: ...
    def __len__(self): ...
    def __iter__(self): ...
    def __getitem__(self, key: Any): ...
    @property
    def column_types(self): ...
    @property
    def column_names(self): ...
    @property
    def row_names(self): ...
    @property
    def columns(self): ...
    @property
    def rows(self): ...
    def print_csv(self, **kwargs: Any) -> None: ...
    def print_json(self, **kwargs: Any) -> None: ...
    def where(self, test: Callable[[Row], bool]) -> "Table": ...
    def select(self, key: Union[Iterable[str], str]) -> "Table": ...
    # these definitions are much narrower than what's actually accepted
    @classmethod
    def from_object(
        cls, obj: Iterable[Dict[str, Any]], *, column_types: Optional["TypeTester"] = None
    ) -> "Table": ...
    @classmethod
    def from_csv(
        cls, path: Iterable[str], *, column_types: Optional["TypeTester"] = None
    ) -> "Table": ...
    @classmethod
    def merge(cls, tables: Iterable["Table"]) -> "Table": ...
    def rename(
        self,
        column_names: Optional[Iterable[str]] = None,
        row_names: Optional[Any] = None,
        slug_columns: bool = False,
        slug_rows: bool = False,
        **kwargs: Any,
    ) -> "Table": ...

class TypeTester:
    def __init__(
        self, force: Any = ..., limit: Optional[Any] = ..., types: Optional[Any] = ...
    ) -> None: ...
    def run(self, rows: Any, column_names: Any): ...

class MaxPrecision:
    def __init__(self, column_name: Any) -> None: ...

# this is not strictly true, but it's all we care about.
def aggregate(self, aggregations: MaxPrecision) -> int: ...



from pathlib import Path

from dbt_common.events.event_manager_client import get_event_manager
from dbt_common.helper_types import WarnErrorOptions, WarnErrorOptionsV2
from dbt_common.invocation import get_invocation_id
from dbt_common.utils.encoding import ForgivingJSONEncoder
from dbt_common.events.base_types import BaseEvent, EventLevel, EventMsg
from dbt_common.events.logger import LoggerConfig, LineFormat
from dbt_common.exceptions import scrub_secrets, env_secrets
from dbt_common.events.types import Note
from functools import partial
import json
import os
import sys
from typing import Any, Callable, Dict, Optional, TextIO, Union
from google.protobuf.json_format import MessageToDict

LOG_VERSION = 3
metadata_vars: Optional[Dict[str, str]] = None
_METADATA_ENV_PREFIX = "DBT_ENV_CUSTOM_ENV_"
WARN_ERROR_OPTIONS: Union[WarnErrorOptions, WarnErrorOptionsV2] = WarnErrorOptions(
    include=[], exclude=[]
)
WARN_ERROR = False

# This global, and the following two functions for capturing stdout logs are
# an unpleasant hack we intend to remove as part of API-ification. The GitHub
# issue #6350 was opened for that work.
CAPTURE_STREAM: Optional[TextIO] = None


def stdout_filter(
    log_cache_events: bool,
    line_format: LineFormat,
    msg: EventMsg,
) -> bool:
    return msg.info.name not in ["CacheAction", "CacheDumpGraph"] or log_cache_events


def get_stdout_config(
    line_format: LineFormat,
    use_colors: bool,
    level: EventLevel,
    log_cache_events: bool,
) -> LoggerConfig:
    return LoggerConfig(
        name="stdout_log",
        level=level,
        use_colors=use_colors,
        line_format=line_format,
        scrubber=env_scrubber,
        filter=partial(
            stdout_filter,
            log_cache_events,
            line_format,
        ),
        invocation_id=get_invocation_id(),
        output_stream=sys.stdout,
    )


def make_log_dir_if_missing(log_path: Union[Path, str]) -> None:
    if isinstance(log_path, str):
        log_path = Path(log_path)
    log_path.mkdir(parents=True, exist_ok=True)


def env_scrubber(msg: str) -> str:
    return scrub_secrets(msg, env_secrets())


# used for integration tests
def capture_stdout_logs(stream: TextIO) -> None:
    global CAPTURE_STREAM
    CAPTURE_STREAM = stream


def stop_capture_stdout_logs() -> None:
    global CAPTURE_STREAM
    CAPTURE_STREAM = None


def get_capture_stream() -> Optional[TextIO]:
    return CAPTURE_STREAM


# returns a dictionary representation of the event fields.
# the message may contain secrets which must be scrubbed at the usage site.
def msg_to_json(msg: EventMsg) -> str:
    msg_dict = msg_to_dict(msg)
    raw_log_line = json.dumps(msg_dict, sort_keys=True, cls=ForgivingJSONEncoder)
    return raw_log_line


def msg_to_dict(msg: EventMsg) -> dict:
    msg_dict = dict()
    try:
        msg_dict = MessageToDict(
            msg,
            preserving_proto_field_name=True,
            always_print_fields_with_no_presence=True,
        )
    except Exception as exc:
        event_type = type(msg).__name__
        fire_event(
            Note(msg=f"type {event_type} is not serializable. {str(exc)}"), level=EventLevel.WARN
        )
    # We don't want an empty NodeInfo in output
    if (
        "data" in msg_dict
        and "node_info" in msg_dict["data"]
        and msg_dict["data"]["node_info"]["node_name"] == ""
    ):
        del msg_dict["data"]["node_info"]
    return msg_dict


# This function continues to exist to provide backwards compatibility
def warn_or_error(event, node=None) -> None:
    fire_event(e=event, node=node, force_warn_or_error_handling=True)


# an alternative to fire_event which only creates and logs the event value
# if the condition is met. Does nothing otherwise.
def fire_event_if(
    conditional: bool, lazy_e: Callable[[], BaseEvent], level: Optional[EventLevel] = None
) -> None:
    if conditional:
        fire_event(lazy_e(), level=level)


# top-level method for accessing the new eventing system
# this is where all the side effects happen branched by event type
# (i.e. - mutating the event history, printing to stdout, logging
# to files, etc.)
def fire_event(
    e: BaseEvent,
    level: Optional[EventLevel] = None,
    node: Any = None,
    force_warn_or_error_handling: bool = False,
) -> None:
    get_event_manager().fire_event(
        e, level=level, node=node, force_warn_or_error_handling=force_warn_or_error_handling
    )


def get_metadata_vars() -> Dict[str, str]:
    global metadata_vars
    if metadata_vars is None:
        metadata_vars = {
            k[len(_METADATA_ENV_PREFIX) :]: v
            for k, v in os.environ.items()
            if k.startswith(_METADATA_ENV_PREFIX)
        }
    return metadata_vars


def reset_metadata_vars() -> None:
    global metadata_vars
    metadata_vars = None



from mashumaro.dialect import Dialect as Dialect
from mashumaro.mixins.dict import DataClassDictMixin as DataClassDictMixin
from typing import Any, Callable, Dict, TypeVar, Union

T = TypeVar("T", bound="DataClassORJSONMixin")
EncodedData = Union[str, bytes, bytearray]
Encoder = Callable[[Any], EncodedData]
Decoder = Callable[[EncodedData], Dict[Any, Any]]

class OrjsonDialect(Dialect):
    serialization_strategy: Any

class DataClassORJSONMixin(DataClassDictMixin):
    def to_jsonb(
        self, encoder: Encoder = ..., *, orjson_options: int = ..., **to_dict_kwargs: Any
    ) -> bytes: ...
    def to_json(
        self, encoder: Encoder = ..., *, orjson_options: int = ..., **to_dict_kwargs: Any
    ) -> bytes: ...
    @classmethod
    def from_json(
        cls, data: EncodedData, decoder: Decoder = ..., **from_dict_kwargs: Any
    ) -> T: ...



import datetime
import decimal
import hashlib
import json
from typing import Tuple, Type, Any

import jinja2
import sys

DECIMALS: Tuple[Type[Any], ...]
try:
    import cdecimal  # type: ignore
except ImportError:
    DECIMALS = (decimal.Decimal,)
else:
    DECIMALS = (decimal.Decimal, cdecimal.Decimal)


def md5(string, charset="utf-8"):
    if sys.version_info >= (3, 9):
        return hashlib.md5(string.encode(charset), usedforsecurity=False).hexdigest()
    else:
        return hashlib.md5(string.encode(charset)).hexdigest()


class JSONEncoder(json.JSONEncoder):
    """A 'custom' json encoder.

    A 'custom' json encoder that does normal json encoder things, but also
    handles `Decimal`s and `Undefined`s. Decimals can lose precision because
    they get converted to floats. Undefined's are serialized to an empty string
    """

    def default(self, obj):
        if isinstance(obj, DECIMALS):
            return float(obj)
        elif isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()
        elif isinstance(obj, jinja2.Undefined):
            return ""
        elif isinstance(obj, Exception):
            return repr(obj)
        elif hasattr(obj, "to_dict"):
            # if we have a to_dict we should try to serialize the result of
            # that!
            return obj.to_dict(omit_none=True)
        else:
            return super().default(obj)


class ForgivingJSONEncoder(JSONEncoder):
    def default(self, obj):
        # let dbt's default JSON encoder handle it if possible, fallback to
        # str()
        try:
            return super().default(obj)
        except TypeError:
            return str(obj)
